import pytest

from darts.logging import get_logger
from darts.metrics import mae
from darts.tests.base_test_class import DartsBaseTestClass
from darts.utils import timeseries_generation as tg

logger = get_logger(__name__)

try:
    import torch

    from darts.models.forecasting.tcn_model import TCNModel

    TORCH_AVAILABLE = True
except ImportError:
    logger.warning("Torch not available. TCN tests will be skipped.")
    TORCH_AVAILABLE = False


if TORCH_AVAILABLE:

    class TCNModelTestCase(DartsBaseTestClass):
        def test_creation(self):
            with self.assertRaises(ValueError):
                # cannot choose a kernel size larger than the input length
                TCNModel(input_chunk_length=20, output_chunk_length=1, kernel_size=100)
            TCNModel(input_chunk_length=12, output_chunk_length=1)

        def test_fit(self):
            large_ts = tg.constant_timeseries(length=100, value=1000)
            small_ts = tg.constant_timeseries(length=100, value=10)

            # Test basic fit and predict
            model = TCNModel(
                input_chunk_length=12, output_chunk_length=1, n_epochs=10, num_layers=1
            )
            model.fit(large_ts[:98])
            pred = model.predict(n=2).values()[0]

            # Test whether model trained on one series is better than one trained on another
            model2 = TCNModel(
                input_chunk_length=12, output_chunk_length=1, n_epochs=10, num_layers=1
            )
            model2.fit(small_ts[:98])
            pred2 = model2.predict(n=2).values()[0]
            self.assertTrue(abs(pred2 - 10) < abs(pred - 10))

            # test short predict
            pred3 = model2.predict(n=1)
            self.assertEqual(len(pred3), 1)

        def test_performance(self):
            # test TCN performance on dummy time series
            ts = tg.sine_timeseries(length=100) + tg.linear_timeseries(
                length=100, end_value=2
            )
            train, test = ts[:90], ts[90:]
            model = TCNModel(
                input_chunk_length=12,
                output_chunk_length=10,
                n_epochs=300,
                random_state=0,
            )
            model.fit(train)
            pred = model.predict(n=10)

            self.assertTrue(mae(pred, test) < 0.3)

        @pytest.mark.slow
        def test_coverage(self):
            torch.manual_seed(0)
            input_chunk_lengths = range(20, 50)
            kernel_sizes = range(2, 5)
            dilation_bases = range(2, 5)

            for kernel_size in kernel_sizes:
                for dilation_base in dilation_bases:
                    if dilation_base > kernel_size:
                        continue
                    for input_chunk_length in input_chunk_lengths:

                        # create model with all weights set to one
                        model = TCNModel(
                            input_chunk_length=input_chunk_length,
                            output_chunk_length=1,
                            kernel_size=kernel_size,
                            dilation_base=dilation_base,
                            weight_norm=False,
                            n_epochs=1,
                        )

                        # we have to fit the model on a dummy series in order to create the internal nn.Module
                        model.fit(tg.gaussian_timeseries(length=100))

                        for res_block in model.model.res_blocks:
                            res_block.conv1.weight = torch.nn.Parameter(
                                torch.ones(
                                    res_block.conv1.weight.shape, dtype=torch.float64
                                )
                            )
                            res_block.conv2.weight = torch.nn.Parameter(
                                torch.ones(
                                    res_block.conv2.weight.shape, dtype=torch.float64
                                )
                            )

                        model.model.eval()

                        # also disable MC Dropout:
                        model.model.set_mc_dropout(False)

                        input_tensor = torch.zeros(
                            [1, input_chunk_length, 1], dtype=torch.float64
                        )
                        zero_output = model.model.forward((input_tensor, None))[
                            0, -1, 0
                        ]

                        # test for full coverage
                        for i in range(input_chunk_length):
                            input_tensor[0, i, 0] = 1
                            curr_output = model.model.forward((input_tensor, None))[
                                0, -1, 0
                            ]
                            self.assertNotEqual(zero_output, curr_output)
                            input_tensor[0, i, 0] = 0

                        # create model with all weights set to one and one layer less than is automatically detected
                        model_2 = TCNModel(
                            input_chunk_length=input_chunk_length,
                            output_chunk_length=1,
                            kernel_size=kernel_size,
                            dilation_base=dilation_base,
                            weight_norm=False,
                            num_layers=model.model.num_layers - 1,
                            n_epochs=1,
                        )

                        # we have to fit the model on a dummy series in order to create the internal nn.Module
                        model_2.fit(tg.gaussian_timeseries(length=100))

                        for res_block in model_2.model.res_blocks:
                            res_block.conv1.weight = torch.nn.Parameter(
                                torch.ones(
                                    res_block.conv1.weight.shape, dtype=torch.float64
                                )
                            )
                            res_block.conv2.weight = torch.nn.Parameter(
                                torch.ones(
                                    res_block.conv2.weight.shape, dtype=torch.float64
                                )
                            )

                        model_2.model.eval()

                        # also disable MC Dropout:
                        model_2.model.set_mc_dropout(False)

                        input_tensor = torch.zeros(
                            [1, input_chunk_length, 1], dtype=torch.float64
                        )
                        zero_output = model_2.model.forward((input_tensor, None))[
                            0, -1, 0
                        ]

                        # test for incomplete coverage
                        uncovered_input_found = False
                        if model_2.model.num_layers == 1:
                            continue
                        for i in range(input_chunk_length):
                            input_tensor[0, i, 0] = 1
                            curr_output = model_2.model.forward((input_tensor, None))[
                                0, -1, 0
                            ]
                            if zero_output == curr_output:
                                uncovered_input_found = True
                                break
                            input_tensor[0, i, 0] = 0
                        self.assertTrue(uncovered_input_found)

        def helper_test_pred_length(self, pytorch_model, series):
            model = pytorch_model(
                input_chunk_length=12, output_chunk_length=3, n_epochs=1
            )
            model.fit(series)
            pred = model.predict(7)
            self.assertEqual(len(pred), 7)
            pred = model.predict(2)
            self.assertEqual(len(pred), 2)
            self.assertEqual(pred.width, 1)
            pred = model.predict(4)
            self.assertEqual(len(pred), 4)
            self.assertEqual(pred.width, 1)

        def test_pred_length(self):
            series = tg.linear_timeseries(length=100)
            self.helper_test_pred_length(TCNModel, series)



"""
Threshold Detector
------------------

Detector that detects anomaly based on user-given threshold.
This detector compares time series values with user-given thresholds, and
identifies time points as anomalous when values are beyond the thresholds.
"""

from typing import Sequence, Union

import numpy as np

from darts.ad.detectors.detectors import Detector
from darts.logging import raise_if, raise_if_not
from darts.timeseries import TimeSeries


class ThresholdDetector(Detector):
    def __init__(
        self,
        low_threshold: Union[int, float, Sequence[float], None] = None,
        high_threshold: Union[int, float, Sequence[float], None] = None,
    ) -> None:
        """
        Flags values that are either below or above the `low_threshold` and `high_threshold`,
        respectively.

        If a single value is provided for `low_threshold` or `high_threshold`, this same
        value will be used across all components of the series.

        If sequences of values are given for the parameters `low_threshold` and/or `high_threshold`,
        they must be of the same length, matching the dimensionality of the series passed
        to ``detect()``, or have a length of 1. In the latter case, this single value will be used
        across all components of the series.

        If either `low_threshold` or `high_threshold` is None, the corresponding bound will not be used.
        However, at least one of the two must be set.

        Parameters
        ----------
        low_threshold
            (Sequence of) lower bounds.
            If a sequence, must match the dimensionality of the series
            this detector is applied to.
        high_threshold
            (Sequence of) upper bounds.
            If a sequence, must match the dimensionality of the series
            this detector is applied to.
        """

        # TODO: could we refactor some code common between ThresholdDetector and QuantileDetector?

        super().__init__()

        raise_if(
            low_threshold is None and high_threshold is None,
            "At least one parameter must be not None (`low` and `high` are both None).",
        )

        def _prep_thresholds(q):
            return (
                q.tolist()
                if isinstance(q, np.ndarray)
                else [q]
                if not isinstance(q, Sequence)
                else q
            )

        low = _prep_thresholds(low_threshold)
        high = _prep_thresholds(high_threshold)

        self.low_threshold = low * len(high) if len(low) == 1 else low
        self.high_threshold = high * len(low) if len(high) == 1 else high

        # the threshold parameters are now sequences of the same length,
        # possibly containing some None values, but at least one non-None value

        raise_if_not(
            len(self.low_threshold) == len(self.high_threshold),
            "Parameters `low_threshold` and `high_threshold` must be of the same length,"
            + f" found `low`: {len(self.low_threshold)} and `high`: {len(self.high_threshold)}.",
        )

        raise_if(
            all([lo is None for lo in self.low_threshold])
            and all([hi is None for hi in self.high_threshold]),
            "All provided threshold values are None.",
        )

        raise_if_not(
            all(
                [
                    l <= h
                    for (l, h) in zip(self.low_threshold, self.high_threshold)
                    if ((l is not None) and (h is not None))
                ]
            ),
            "all values in `low_threshold` must be lower than or equal"
            + "to their corresponding value in `high_threshold`.",
        )

    def _detect_core(self, series: TimeSeries) -> TimeSeries:
        raise_if_not(
            series.is_deterministic, "This detector only works on deterministic series."
        )

        raise_if(
            len(self.low_threshold) > 1 and len(self.low_threshold) != series.width,
            "The number of components of input must be equal to the number"
            + " of threshold values. Found number of "
            + f"components equal to {series.width} and expected {len(self.low_threshold)}.",
        )

        # if length is 1, tile it to series width:
        low_threshold = (
            self.low_threshold * series.width
            if len(self.low_threshold) == 1
            else self.low_threshold
        )
        high_threshold = (
            self.high_threshold * series.width
            if len(self.high_threshold) == 1
            else self.high_threshold
        )

        # (time, components)
        np_series = series.all_values(copy=False).squeeze(-1)

        def _detect_fn(x, lo, hi):
            # x of shape (time,) for 1 component
            return (x < (np.NINF if lo is None else lo)) | (
                x > (np.Inf if hi is None else hi)
            )

        detected = np.zeros_like(np_series, dtype=int)

        for component_idx in range(series.width):
            detected[:, component_idx] = _detect_fn(
                np_series[:, component_idx],
                low_threshold[component_idx],
                high_threshold[component_idx],
            )

        return TimeSeries.from_times_and_values(series.time_index, detected)



"""
Forecasting Anomaly Model
-------------------------

A ``ForecastingAnomalyModel`` wraps around a Darts forecasting model and one or several anomaly
scorer(s) to compute anomaly scores by comparing how actuals deviate from the model's forecasts.
"""

# TODO:
#     - put start default value to its minimal value (wait for the release of historical_forecast)

import inspect
from typing import Dict, Optional, Sequence, Union

import pandas as pd

from darts.ad.anomaly_model.anomaly_model import AnomalyModel
from darts.ad.scorers.scorers import AnomalyScorer
from darts.ad.utils import _assert_same_length, _assert_timeseries, _to_list
from darts.logging import get_logger, raise_if_not
from darts.models.forecasting.forecasting_model import ForecastingModel
from darts.timeseries import TimeSeries

logger = get_logger(__name__)


class ForecastingAnomalyModel(AnomalyModel):
    def __init__(
        self,
        model: ForecastingModel,
        scorer: Union[AnomalyScorer, Sequence[AnomalyScorer]],
    ):
        """Forecasting-based Anomaly Detection Model

        The forecasting model may or may not be already fitted. The underlying assumption is that `model`
        should be able to accurately forecast the series in the absence of anomalies. For this reason,
        it is recommended to either provide a model that has already been fitted and evaluated to work
        appropriately on a series without anomalies, or to ensure that a simple call to the :func:`fit()`
        method of the model will be sufficient to train it to satisfactory performance on a series without anomalies.

        Calling :func:`fit()` on the anomaly model will fit the underlying forecasting model only
        if ``allow_model_training`` is set to ``True`` upon calling ``fit()``.
        In addition, calling :func:`fit()` will also fit the fittable scorers, if any.

        Parameters
        ----------
        model
            An instance of a Darts forecasting model.
        scorer
            One or multiple scorer(s) that will be used to compare the actual and predicted time series in order
            to obtain an anomaly score ``TimeSeries``.
            If a list of `N` scorers is given, the anomaly model will call each
            one of the scorers and output a list of `N` anomaly scores ``TimeSeries``.
        """

        raise_if_not(
            isinstance(model, ForecastingModel),
            f"Model must be a darts ForecastingModel not a {type(model)}.",
        )
        self.model = model

        super().__init__(model=model, scorer=scorer)

    def fit(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        allow_model_training: bool = False,
        forecast_horizon: int = 1,
        start: Union[pd.Timestamp, float, int] = 0.5,
        num_samples: int = 1,
        **model_fit_kwargs,
    ):
        """Fit the underlying forecasting model (if applicable) and the fittable scorers, if any.

        Train the model (if not already fitted and ``allow_model_training`` is set to True) and the
        scorer(s) (if fittable) on the given time series.

        Once the model is fitted, the series historical forecasts are computed,
        representing what would have been forecasted by this model on the series.

        The prediction and the series are then used to train the scorer(s).

        Parameters
        ----------
        series
            One or multiple (if the model supports it) target series to be
            trained on (generally assumed to be anomaly-free).
        past_covariates
            Optional past-observed covariate series or sequence of series. This applies only if the model
            supports past covariates.
        future_covariates
            Optional future-known covariate series or sequence of series. This applies only if the model
            supports future covariates.
        allow_model_training
            Boolean value that indicates if the forecasting model needs to be fitted on the given series.
            If set to False, the model needs to be already fitted.
            Default: False
        forecast_horizon
            The forecast horizon for the predictions.
        start
            The first point of time at which a prediction is computed for a future time.
            This parameter supports 3 different data types: ``float``, ``int`` and ``pandas.Timestamp``.
            In the case of ``float``, the parameter will be treated as the proportion of the time series
            that should lie before the first prediction point.
            In the case of ``int``, the parameter will be treated as an integer index to the time index of
            `series` that will be used as first prediction time.
            In case of ``pandas.Timestamp``, this time stamp will be used to determine the first prediction time
            directly.
            Default: 0.5
        num_samples
            Number of times a prediction is sampled from a probabilistic model. Should be left set to 1 for
            deterministic models.
        model_fit_kwargs
            Parameters to be passed on to the forecast model ``fit()`` method.

        Returns
        -------
        self
            Fitted model
        """

        raise_if_not(
            type(allow_model_training) is bool,
            f"`allow_model_training` must be Boolean, found type: {type(allow_model_training)}.",
        )

        # checks if model does not need training and all scorer(s) are not fittable
        if not allow_model_training and not self.scorers_are_trainable:
            logger.warning(
                f"The forecasting model {self.model.__class__.__name__} won't be trained"
                + " because the parameter `allow_model_training` is set to False, and no scorer"
                + " is fittable. ``.fit()`` method has no effect."
            )
            return

        list_series = _to_list(series)

        raise_if_not(
            all([isinstance(s, TimeSeries) for s in list_series]),
            "all input `series` must be of type Timeseries.",
        )

        list_past_covariates = self._prepare_covariates(
            past_covariates, list_series, "past"
        )
        list_future_covariates = self._prepare_covariates(
            future_covariates, list_series, "future"
        )

        model_fit_kwargs["past_covariates"] = list_past_covariates
        model_fit_kwargs["future_covariates"] = list_future_covariates

        # remove None elements from dictionary
        model_fit_kwargs = {k: v for k, v in model_fit_kwargs.items() if v}

        # fit forecasting model
        if allow_model_training:
            # the model has not been trained yet

            fit_signature_series = (
                inspect.signature(self.model.fit).parameters["series"].annotation
            )

            # checks if model can be trained on multiple time series or only on a time series
            # TODO: check if model can accept multivariate timeseries, raise error if given and model cannot
            if "Sequence[darts.timeseries.TimeSeries]" in str(fit_signature_series):
                self.model.fit(series=list_series, **model_fit_kwargs)
            else:
                raise_if_not(
                    len(list_series) == 1,
                    f"Forecasting model {self.model.__class__.__name__} only accepts a single time series"
                    + " for the training phase and not a sequence of multiple of time series.",
                )
                self.model.fit(series=list_series[0], **model_fit_kwargs)
        else:
            raise_if_not(
                self.model._fit_called,
                f"Model {self.model.__class__.__name__} needs to be trained, consider training "
                + "it beforehand or setting "
                + "`allow_model_training` to True (default: False). "
                + "The model will then be trained on the provided series.",
            )

        # generate the historical_forecast() prediction of the model on the train set
        if self.scorers_are_trainable:
            # check if the window size of the scorers are lower than the max size allowed
            self._check_window_size(list_series, start)

            list_pred = []
            for idx, series in enumerate(list_series):

                if list_past_covariates is not None:
                    past_covariates = list_past_covariates[idx]

                if list_future_covariates is not None:
                    future_covariates = list_future_covariates[idx]

                list_pred.append(
                    self._predict_with_forecasting(
                        series,
                        past_covariates=past_covariates,
                        future_covariates=future_covariates,
                        forecast_horizon=forecast_horizon,
                        start=start,
                        num_samples=num_samples,
                    )
                )

        # fit the scorers
        for scorer in self.scorers:
            if hasattr(scorer, "fit"):
                scorer.fit_from_prediction(list_series, list_pred)

        return self

    def _prepare_covariates(
        self,
        covariates: Union[TimeSeries, Sequence[TimeSeries]],
        series: Sequence[TimeSeries],
        name_covariates: str,
    ) -> Sequence[TimeSeries]:
        """Convert `covariates` into Sequence, if not already, and checks if their length is equal to the one of `series`.

        Parameters
        ----------
        covariates
            Covariate ("future" or "past") of `series`.
        series
            The series to be trained on.
        name_covariates
            Internal parameter for error message, a string indicating if it is a "future" or "past" covariates.

        Returns
        -------
        Sequence[TimeSeries]
            Covariate time series
        """

        if covariates is not None:
            list_covariates = _to_list(covariates)

            for covariates in list_covariates:
                _assert_timeseries(
                    covariates, name_covariates + "_covariates input series"
                )

            raise_if_not(
                len(list_covariates) == len(series),
                f"Number of {name_covariates}_covariates must match the number of given "
                + f"series, found length {len(list_covariates)} and expected {len(series)}.",
            )

        return list_covariates if covariates is not None else None

    def show_anomalies(
        self,
        series: TimeSeries,
        past_covariates: Optional[TimeSeries] = None,
        future_covariates: Optional[TimeSeries] = None,
        forecast_horizon: int = 1,
        start: Union[pd.Timestamp, float, int] = 0.5,
        num_samples: int = 1,
        actual_anomalies: TimeSeries = None,
        names_of_scorers: Union[str, Sequence[str]] = None,
        title: str = None,
        metric: str = None,
    ):
        """Plot the results of the anomaly model.

        Computes the score on the given series input and shows the different anomaly scores with respect to time.

        The plot will be composed of the following:

        - the series itself with the output of the forecasting model.
        - the anomaly score for each scorer. The scorers with different windows will be separated.
        - the actual anomalies, if given.

        It is possible to:

        - add a title to the figure with the parameter `title`
        - give personalized names for the scorers with `names_of_scorers`
        - show the results of a metric for each anomaly score (AUC_ROC or AUC_PR),
            if the actual anomalies are provided.

        Parameters
        ----------
        series
            The series to visualize anomalies from.
        past_covariates
            An optional past-observed covariate series or sequence of series. This applies only if the model
            supports past covariates.
        future_covariates
            An optional future-known covariate series or sequence of series. This applies only if the model
            supports future covariates.
        forecast_horizon
            The forecast horizon for the predictions.
        start
            The first point of time at which a prediction is computed for a future time.
            This parameter supports 3 different data types: ``float``, ``int`` and ``pandas.Timestamp``.
            In the case of ``float``, the parameter will be treated as the proportion of the time series
            that should lie before the first prediction point.
            In the case of ``int``, the parameter will be treated as an integer index to the time index of
            `series` that will be used as first prediction time.
            In case of ``pandas.Timestamp``, this time stamp will be used to determine the first prediction time
            directly.
        num_samples
            Number of times a prediction is sampled from a probabilistic model. Should be left set to 1 for
            deterministic models.
        actual_anomalies
            The ground truth of the anomalies (1 if it is an anomaly and 0 if not)
        names_of_scorers
            Name of the scores. Must be a list of length equal to the number of scorers in the anomaly_model.
        title
            Title of the figure
        metric
            Optionally, Scoring function to use. Must be one of "AUC_ROC" and "AUC_PR".
            Default: "AUC_ROC"
        """

        if isinstance(series, Sequence):
            raise_if_not(
                len(series) == 1,
                f"`show_anomalies` expects one series, found a list of length {len(series)} as input.",
            )

            series = series[0]

        raise_if_not(
            isinstance(series, TimeSeries),
            f"`show_anomalies` expects an input of type TimeSeries, found type: {type(series)}.",
        )

        anomaly_scores, model_output = self.score(
            series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
            forecast_horizon=forecast_horizon,
            start=start,
            num_samples=num_samples,
            return_model_prediction=True,
        )

        return self._show_anomalies(
            series,
            model_output=model_output,
            anomaly_scores=anomaly_scores,
            names_of_scorers=names_of_scorers,
            actual_anomalies=actual_anomalies,
            title=title,
            metric=metric,
        )

    def score(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        forecast_horizon: int = 1,
        start: Union[pd.Timestamp, float, int] = 0.5,
        num_samples: int = 1,
        return_model_prediction: bool = False,
    ) -> Union[TimeSeries, Sequence[TimeSeries]]:
        """Compute anomaly score(s) for the given series.

        Predicts the given target time series with the forecasting model, and applies the scorer(s)
        on the prediction and the target input time series. Outputs the anomaly score of the given
        input time series.

        Parameters
        ----------
        series
            The (sequence of) series to score on.
        past_covariates
            An optional past-observed covariate series or sequence of series. This applies only if the model
            supports past covariates.
        future_covariates
            An optional future-known covariate series or sequence of series. This applies only if the model
            supports future covariates.
        forecast_horizon
            The forecast horizon for the predictions.
        start
            The first point of time at which a prediction is computed for a future time.
            This parameter supports 3 different data types: ``float``, ``int`` and ``pandas.Timestamp``.
            In the case of ``float``, the parameter will be treated as the proportion of the time series
            that should lie before the first prediction point.
            In the case of ``int``, the parameter will be treated as an integer index to the time index of
            `series` that will be used as first prediction time.
            In case of ``pandas.Timestamp``, this time stamp will be used to determine the first prediction time
            directly. Default: 0.5
        num_samples
            Number of times a prediction is sampled from a probabilistic model. Should be left set to 1 for
            deterministic models.
        return_model_prediction
            Boolean value indicating if the prediction of the model should be returned along the anomaly score
            Default: False

        Returns
        -------
        Union[TimeSeries, Sequence[TimeSeries], Sequence[Sequence[TimeSeries]]]
            Anomaly scores series generated by the anomaly model scorers

                - ``TimeSeries`` if `series` is a series, and the anomaly model contains one scorer.
                - ``Sequence[TimeSeries]``

                    * if `series` is a series, and the anomaly model contains multiple scorers,
                      returns one series per scorer.
                    * if `series` is a sequence, and the anomaly model contains one scorer,
                      returns one series per series in the sequence.
                - ``Sequence[Sequence[TimeSeries]]`` if `series` is a sequence, and the anomaly
                  model contains multiple scorers. The outer sequence is over the series,
                  and inner sequence is over the scorers.
        """
        raise_if_not(
            type(return_model_prediction) is bool,
            f"`return_model_prediction` must be Boolean, found type: {type(return_model_prediction)}.",
        )

        raise_if_not(
            self.model._fit_called,
            f"Model {self.model} has not been trained. Please call ``.fit()``.",
        )

        list_series = _to_list(series)

        list_past_covariates = self._prepare_covariates(
            past_covariates, list_series, "past"
        )
        list_future_covariates = self._prepare_covariates(
            future_covariates, list_series, "future"
        )

        # check if the window size of the scorers are lower than the max size allowed
        self._check_window_size(list_series, start)

        list_pred = []
        for idx, s in enumerate(list_series):

            if list_past_covariates is not None:
                past_covariates = list_past_covariates[idx]

            if list_future_covariates is not None:
                future_covariates = list_future_covariates[idx]

            list_pred.append(
                self._predict_with_forecasting(
                    s,
                    past_covariates=past_covariates,
                    future_covariates=future_covariates,
                    forecast_horizon=forecast_horizon,
                    start=start,
                    num_samples=num_samples,
                )
            )

        scores = list(
            zip(
                *[
                    sc.score_from_prediction(list_series, list_pred)
                    for sc in self.scorers
                ]
            )
        )

        if len(scores) == 1 and not isinstance(series, Sequence):
            # there's only one series
            scores = scores[0]
            if len(scores) == 1:
                # there's only one scorer
                scores = scores[0]

        if len(list_pred) == 1:
            list_pred = list_pred[0]

        if return_model_prediction:
            return scores, list_pred
        else:
            return scores

    def _check_window_size(
        self, series: Sequence[TimeSeries], start: Union[pd.Timestamp, float, int]
    ):
        """Checks if the parameters `window` of the scorers are smaller than the maximum window size allowed.
        The maximum size allowed is equal to the output length of the .historical_forecast() applied on `series`.
        It is defined by the parameter `start` and the series’ length.

        Parameters
        ----------
        series
            The series given to the .historical_forecast()
        start
            Parameter of the .historical_forecast(): first point of time at which a prediction is computed
            for a future time.
        """
        # biggest window of the anomaly_model scorers
        max_window = max(scorer.window for scorer in self.scorers)

        for s in series:
            max_possible_window = (
                len(s.drop_before(s.get_timestamp_at_point(start))) + 1
            )
            raise_if_not(
                max_window <= max_possible_window,
                f"Window size {max_window} is greater than the targeted series length {max_possible_window},"
                + f" must be lower or equal. Reduce window size, or reduce start value (start: {start}).",
            )

    def _predict_with_forecasting(
        self,
        series: TimeSeries,
        past_covariates: Optional[TimeSeries] = None,
        future_covariates: Optional[TimeSeries] = None,
        forecast_horizon: int = 1,
        start: Union[pd.Timestamp, float, int] = None,
        num_samples: int = 1,
    ) -> TimeSeries:

        """Compute the historical forecasts that would have been obtained by this model on the `series`.

        `retrain` is set to False if possible (this is not supported by all models). If set to True, it will always
        re-train the model on the entire available history,

        Parameters
        ----------
        series
            The target time series to use to successively train and evaluate the historical forecasts.
        past_covariates
            An optional past-observed covariate series or sequence of series. This applies only if the model
            supports past covariates.
        future_covariates
            An optional future-known covariate series or sequence of series. This applies only if the model
            supports future covariates.
        forecast_horizon
            The forecast horizon for the predictions
        start
            The first point of time at which a prediction is computed for a future time.
            This parameter supports 3 different data types: ``float``, ``int`` and ``pandas.Timestamp``.
            In the case of ``float``, the parameter will be treated as the proportion of the time series
            that should lie before the first prediction point.
            In the case of ``int``, the parameter will be treated as an integer index to the time index of
            `series` that will be used as first prediction time.
            In case of ``pandas.Timestamp``, this time stamp will be used to determine the first prediction time
            directly.
        num_samples
            Number of times a prediction is sampled from a probabilistic model. Should be left set to 1 for
            deterministic models.

        Returns
        -------
        TimeSeries
            Single ``TimeSeries`` instance created from the last point of each individual forecast.
        """

        # TODO: raise an exception. We only support models that do not need retrain
        # checks if model accepts to not be retrained in the historical_forecasts()
        if self.model._supports_non_retrainable_historical_forecasts():
            # default: set to False. Allows a faster computation.
            retrain = False
        else:
            retrain = True

        historical_forecasts_param = {
            "past_covariates": past_covariates,
            "future_covariates": future_covariates,
            "forecast_horizon": forecast_horizon,
            "start": start,
            "retrain": retrain,
            "num_samples": num_samples,
            "stride": 1,
            "last_points_only": True,
            "verbose": False,
        }

        return self.model.historical_forecasts(series, **historical_forecasts_param)

    def eval_accuracy(
        self,
        actual_anomalies: Union[TimeSeries, Sequence[TimeSeries]],
        series: Union[TimeSeries, Sequence[TimeSeries]],
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        forecast_horizon: int = 1,
        start: Union[pd.Timestamp, float, int] = 0.5,
        num_samples: int = 1,
        metric: str = "AUC_ROC",
    ) -> Union[
        Dict[str, float],
        Dict[str, Sequence[float]],
        Sequence[Dict[str, float]],
        Sequence[Dict[str, Sequence[float]]],
    ]:
        """Compute the accuracy of the anomaly scores computed by the model.

        Predicts the `series` with the forecasting model, and applies the
        scorer(s) on the predicted time series and the given target time series. Returns the
        score(s) of an agnostic threshold metric, based on the anomaly score given by the scorer(s).

        Parameters
        ----------
        actual_anomalies
            The (sequence of) ground truth of the anomalies (1 if it is an anomaly and 0 if not)
        series
            The (sequence of) series to predict anomalies on.
        past_covariates
            An optional past-observed covariate series or sequence of series. This applies only
            if the model supports past covariates.
        future_covariates
            An optional future-known covariate series or sequence of series. This applies only
            if the model supports future covariates.
        forecast_horizon
            The forecast horizon for the predictions.
        start
            The first point of time at which a prediction is computed for a future time.
            This parameter supports 3 different data types: ``float``, ``int`` and ``pandas.Timestamp``.
            In the case of ``float``, the parameter will be treated as the proportion of the time series
            that should lie before the first prediction point.
            In the case of ``int``, the parameter will be treated as an integer index to the time index of
            `series` that will be used as first prediction time.
            In case of ``pandas.Timestamp``, this time stamp will be used to determine the first prediction time
            directly.
        num_samples
            Number of times a prediction is sampled from a probabilistic model. Should be left set to 1 for
            deterministic models.
        metric
            Optionally, Scoring function to use. Must be one of "AUC_ROC" and "AUC_PR".
            Default: "AUC_ROC"

        Returns
        -------
        Union[Dict[str, float], Dict[str, Sequence[float]], Sequence[Dict[str, float]],
        Sequence[Dict[str, Sequence[float]]]]
            Score for the time series.
            A (sequence of) dictionary with the keys being the name of the scorers, and the values being the
            metric results on the (sequence of) `series`. If the scorer treats every dimension independently
            (by nature of the scorer or if its component_wise is set to True), the values of the dictionary
            will be a Sequence containing the score for each dimension.
        """

        list_actual_anomalies = _to_list(actual_anomalies)
        list_series = _to_list(series)

        raise_if_not(
            all([isinstance(s, TimeSeries) for s in list_series]),
            "all input `series` must be of type Timeseries.",
        )

        raise_if_not(
            all([isinstance(s, TimeSeries) for s in list_actual_anomalies]),
            "all input `actual_anomalies` must be of type Timeseries.",
        )

        _assert_same_length(list_actual_anomalies, list_series)
        self._check_univariate(list_actual_anomalies)

        list_anomaly_scores = self.score(
            series=list_series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
            forecast_horizon=forecast_horizon,
            start=start,
            num_samples=num_samples,
        )

        acc_anomaly_scores = self._eval_accuracy_from_scores(
            list_actual_anomalies=list_actual_anomalies,
            list_anomaly_scores=list_anomaly_scores,
            metric=metric,
        )

        if len(acc_anomaly_scores) == 1 and not isinstance(series, Sequence):
            return acc_anomaly_scores[0]
        else:
            return acc_anomaly_scores



import unittest

import numpy as np
import pandas as pd

from darts import TimeSeries
from darts.logging import get_logger
from darts.models import (
    ExponentialSmoothing,
    LinearRegressionModel,
    NaiveDrift,
    NaiveEnsembleModel,
    NaiveSeasonal,
    RegressionEnsembleModel,
    Theta,
)
from darts.tests.base_test_class import DartsBaseTestClass
from darts.utils import timeseries_generation as tg

logger = get_logger(__name__)

try:
    from darts.models import NBEATSModel, RNNModel, TCNModel

    TORCH_AVAILABLE = True
except ImportError:
    logger.warning("Torch not installed - Some ensemble models tests will be skipped.")
    TORCH_AVAILABLE = False


def _make_ts(start_value=0, n=100):
    times = pd.date_range(start="1/1/2013", periods=n, freq="D")
    pd_series = pd.Series(range(start_value, start_value + n), index=times)
    return TimeSeries.from_series(pd_series)


class EnsembleModelsTestCase(DartsBaseTestClass):
    series1 = tg.sine_timeseries(value_frequency=(1 / 5), value_y_offset=10, length=50)
    series2 = tg.linear_timeseries(length=50)

    seq1 = [_make_ts(0), _make_ts(10), _make_ts(20)]
    cov1 = [_make_ts(5), _make_ts(15), _make_ts(25)]

    def test_untrained_models(self):
        model = NaiveDrift()
        _ = NaiveEnsembleModel([model])

        # trained models should raise error
        model.fit(self.series1)
        with self.assertRaises(ValueError):
            NaiveEnsembleModel([model])

        # an untrained ensemble model should also give untrained underlying models
        model_ens = NaiveEnsembleModel([NaiveDrift()])
        model_ens.fit(self.series1)
        assert model_ens.models[0]._fit_called
        new_model = model_ens.untrained_model()
        assert not new_model.models[0]._fit_called

    def test_extreme_lag_inference(self):
        ensemble = NaiveEnsembleModel([NaiveDrift()])
        assert ensemble.extreme_lags == (
            -3,
            -1,
            None,
            None,
            None,
            None,
        )  # test if default is okay

        model1 = LinearRegressionModel(
            lags=3, lags_past_covariates=[-3, -5], lags_future_covariates=[7, 8]
        )
        model2 = LinearRegressionModel(
            lags=5, lags_past_covariates=6, lags_future_covariates=[6, 9]
        )

        ensemble = NaiveEnsembleModel(
            [model1, model2]
        )  # test if infers extreme lags is okay
        expected = (-5, 0, -6, -1, 6, 9)
        assert expected == ensemble.extreme_lags

    def test_input_models_local_models(self):
        with self.assertRaises(ValueError):
            NaiveEnsembleModel([])
        with self.assertRaises(ValueError):
            NaiveEnsembleModel([NaiveDrift, NaiveSeasonal, Theta, ExponentialSmoothing])
        with self.assertRaises(ValueError):
            NaiveEnsembleModel(
                [NaiveDrift(), NaiveSeasonal, Theta(), ExponentialSmoothing()]
            )
        NaiveEnsembleModel(
            [NaiveDrift(), NaiveSeasonal(), Theta(), ExponentialSmoothing()]
        )

    def test_call_predict_local_models(self):
        naive_ensemble = NaiveEnsembleModel([NaiveSeasonal(), Theta()])
        with self.assertRaises(Exception):
            naive_ensemble.predict(5)
        naive_ensemble.fit(self.series1)
        pred1 = naive_ensemble.predict(5)
        assert self.series1.components == pred1.components

    def test_call_backtest_naive_ensemble_local_models(self):
        ensemble = NaiveEnsembleModel([NaiveSeasonal(5), Theta(2, 5)])
        ensemble.fit(self.series1)
        assert ensemble.extreme_lags == (-10, 0, None, None, None, None)
        ensemble.backtest(self.series1)

    def test_predict_ensemble_local_models(self):
        naive = NaiveSeasonal(K=5)
        theta = Theta()
        naive_ensemble = NaiveEnsembleModel([naive, theta])
        naive_ensemble.fit(self.series1 + self.series2)
        forecast_naive_ensemble = naive_ensemble.predict(5)
        naive.fit(self.series1 + self.series2)
        theta.fit(self.series1 + self.series2)
        forecast_mean = 0.5 * naive.predict(5) + 0.5 * theta.predict(5)

        self.assertTrue(
            np.array_equal(forecast_naive_ensemble.values(), forecast_mean.values())
        )

    def test_stochastic_ensemble(self):
        model1 = LinearRegressionModel(lags=1, likelihood="quantile")
        model2 = LinearRegressionModel(lags=2, likelihood="quantile")

        naive_ensemble = NaiveEnsembleModel([model1, model2])
        self.assertTrue(naive_ensemble._is_probabilistic())

        regression_ensemble = RegressionEnsembleModel(
            [model1, model2], regression_train_n_points=1
        )
        self.assertTrue(regression_ensemble._is_probabilistic())

    @unittest.skipUnless(TORCH_AVAILABLE, "requires torch")
    def test_input_models_global_models(self):
        NaiveEnsembleModel([RNNModel(12), TCNModel(10, 2), NBEATSModel(10, 2)])

    @unittest.skipUnless(TORCH_AVAILABLE, "requires torch")
    def test_call_predict_global_models_univariate_input_no_covariates(self):
        naive_ensemble = NaiveEnsembleModel(
            [
                RNNModel(12, n_epochs=1),
                TCNModel(10, 2, n_epochs=1),
                NBEATSModel(10, 2, n_epochs=1),
            ]
        )
        with self.assertRaises(Exception):
            naive_ensemble.predict(5)

        naive_ensemble.fit(self.series1)
        naive_ensemble.predict(5)

    @unittest.skipUnless(TORCH_AVAILABLE, "requires torch")
    def test_call_predict_global_models_multivariate_input_no_covariates(self):
        naive_ensemble = NaiveEnsembleModel(
            [
                RNNModel(12, n_epochs=1),
                TCNModel(10, 2, n_epochs=1),
                NBEATSModel(10, 2, n_epochs=1),
            ]
        )
        naive_ensemble.fit(self.seq1)
        naive_ensemble.predict(n=5, series=self.seq1)

    @unittest.skipUnless(TORCH_AVAILABLE, "requires torch")
    def test_call_predict_global_models_multivariate_input_with_covariates(self):
        naive_ensemble = NaiveEnsembleModel(
            [
                RNNModel(12, n_epochs=1),
                TCNModel(10, 2, n_epochs=1),
                NBEATSModel(10, 2, n_epochs=1),
            ]
        )
        naive_ensemble.fit(self.seq1, self.cov1)
        predict_series = [s[:12] for s in self.seq1]
        predict_covariates = [c[:14] for c in self.cov1]
        naive_ensemble.predict(
            n=2, series=predict_series, past_covariates=predict_covariates
        )

    @unittest.skipUnless(TORCH_AVAILABLE, "requires torch")
    def test_input_models_mixed(self):
        with self.assertRaises(ValueError):
            NaiveEnsembleModel([NaiveDrift(), Theta(), RNNModel(12)])

    def test_fit_multivar_ts_with_local_models(self):
        naive = NaiveEnsembleModel(
            [NaiveDrift(), NaiveSeasonal(), Theta(), ExponentialSmoothing()]
        )
        with self.assertRaises(ValueError):
            naive.fit(self.seq1)

    def test_fit_univar_ts_with_covariates_for_local_models(self):
        naive = NaiveEnsembleModel(
            [NaiveDrift(), NaiveSeasonal(), Theta(), ExponentialSmoothing()]
        )
        with self.assertRaises(ValueError):
            naive.fit(self.series1, self.series2)

    def test_predict_with_target(self):
        series_long = self.series1
        series_short = series_long[:25]

        # train with a single series
        ensemble_model = self.get_global_ensembe_model()
        ensemble_model.fit(series_short, past_covariates=series_long)
        # predict after end of train series
        preds = ensemble_model.predict(n=5, past_covariates=series_long)
        self.assertTrue(isinstance(preds, TimeSeries))
        # predict a new target series
        preds = ensemble_model.predict(
            n=5, series=series_long, past_covariates=series_long
        )
        self.assertTrue(isinstance(preds, TimeSeries))
        # predict multiple target series
        preds = ensemble_model.predict(
            n=5, series=[series_long] * 2, past_covariates=[series_long] * 2
        )
        self.assertTrue(isinstance(preds, list) and len(preds) == 2)
        # predict single target series in list
        preds = ensemble_model.predict(
            n=5, series=[series_long], past_covariates=[series_long]
        )
        self.assertTrue(isinstance(preds, list) and len(preds) == 1)

        # train with multiple series
        ensemble_model = self.get_global_ensembe_model()
        ensemble_model.fit([series_short] * 2, past_covariates=[series_long] * 2)
        with self.assertRaises(ValueError):
            # predict without passing series should raise an error
            ensemble_model.predict(n=5, past_covariates=series_long)
        # predict a new target series
        preds = ensemble_model.predict(
            n=5, series=series_long, past_covariates=series_long
        )
        self.assertTrue(isinstance(preds, TimeSeries))
        # predict multiple target series
        preds = ensemble_model.predict(
            n=5, series=[series_long] * 2, past_covariates=[series_long] * 2
        )
        self.assertTrue(isinstance(preds, list) and len(preds) == 2)
        # predict single target series in list
        preds = ensemble_model.predict(
            n=5, series=[series_long], past_covariates=[series_long]
        )
        self.assertTrue(isinstance(preds, list) and len(preds) == 1)

    @staticmethod
    def get_global_ensembe_model(output_chunk_length=5):
        lags = [-1, -2, -5]
        return NaiveEnsembleModel(
            models=[
                LinearRegressionModel(
                    lags=lags,
                    lags_past_covariates=lags,
                    output_chunk_length=output_chunk_length,
                ),
                LinearRegressionModel(
                    lags=lags,
                    lags_past_covariates=lags,
                    output_chunk_length=output_chunk_length,
                ),
            ],
        )


if __name__ == "__main__":
    import unittest

    unittest.main()



from typing import Union

import numpy as np
import pandas as pd

from darts.tests.base_test_class import DartsBaseTestClass
from darts.utils.timeseries_generation import (
    autoregressive_timeseries,
    constant_timeseries,
    gaussian_timeseries,
    generate_index,
    holidays_timeseries,
    linear_timeseries,
    random_walk_timeseries,
    sine_timeseries,
)


class TimeSeriesGenerationTestCase(DartsBaseTestClass):
    def test_constant_timeseries(self):
        # testing parameters
        value = 5

        def test_routine(start, end=None, length=None):
            # testing for constant value
            constant_ts = constant_timeseries(
                start=start, end=end, value=value, length=length
            )
            value_set = set(constant_ts.values().flatten())
            self.assertTrue(len(value_set) == 1)
            self.assertEqual(len(constant_ts), length_assert)

        for length_assert in [1, 2, 5, 10, 100]:
            test_routine(start=0, length=length_assert)
            test_routine(start=0, end=length_assert - 1)
            test_routine(start=pd.Timestamp("2000-01-01"), length=length_assert)
            end_date = generate_index(
                start=pd.Timestamp("2000-01-01"), length=length_assert
            )[-1]
            test_routine(start=pd.Timestamp("2000-01-01"), end=end_date)

    def test_linear_timeseries(self):

        # testing parameters
        start_value = 5
        end_value = 12

        def test_routine(start, end=None, length=None):
            # testing for start value, end value and delta between two adjacent entries
            linear_ts = linear_timeseries(
                start=start,
                end=end,
                length=length,
                start_value=start_value,
                end_value=end_value,
            )
            self.assertEqual(linear_ts.values()[0][0], start_value)
            self.assertEqual(linear_ts.values()[-1][0], end_value)
            self.assertAlmostEqual(
                linear_ts.values()[-1][0] - linear_ts.values()[-2][0],
                (end_value - start_value) / (length_assert - 1),
            )
            self.assertEqual(len(linear_ts), length_assert)

        for length_assert in [2, 5, 10, 100]:
            test_routine(start=0, length=length_assert)
            test_routine(start=0, end=length_assert - 1)
            test_routine(start=pd.Timestamp("2000-01-01"), length=length_assert)
            end_date = generate_index(
                start=pd.Timestamp("2000-01-01"), length=length_assert
            )[-1]
            test_routine(start=pd.Timestamp("2000-01-01"), end=end_date)

    def test_sine_timeseries(self):

        # testing parameters
        value_amplitude = 5
        value_y_offset = -3

        def test_routine(start, end=None, length=None):
            # testing for correct value range
            sine_ts = sine_timeseries(
                start=start,
                end=end,
                length=length,
                value_amplitude=value_amplitude,
                value_y_offset=value_y_offset,
            )
            self.assertTrue((sine_ts <= value_y_offset + value_amplitude).all().all())
            self.assertTrue((sine_ts >= value_y_offset - value_amplitude).all().all())
            self.assertEqual(len(sine_ts), length_assert)

        for length_assert in [1, 2, 5, 10, 100]:
            test_routine(start=0, length=length_assert)
            test_routine(start=0, end=length_assert - 1)
            test_routine(start=pd.Timestamp("2000-01-01"), length=length_assert)
            end_date = generate_index(
                start=pd.Timestamp("2000-01-01"), length=length_assert
            )[-1]
            test_routine(start=pd.Timestamp("2000-01-01"), end=end_date)

    def test_gaussian_timeseries(self):

        # testing for correct length
        def test_routine(start, end=None, length=None):
            gaussian_ts = gaussian_timeseries(start=start, end=end, length=length)
            self.assertEqual(len(gaussian_ts), length_assert)

        for length_assert in [1, 2, 5, 10, 100]:
            test_routine(start=0, length=length_assert)
            test_routine(start=0, end=length_assert - 1)
            test_routine(start=pd.Timestamp("2000-01-01"), length=length_assert)
            end_date = generate_index(
                start=pd.Timestamp("2000-01-01"), length=length_assert
            )[-1]
            test_routine(start=pd.Timestamp("2000-01-01"), end=end_date)

    def test_random_walk_timeseries(self):

        # testing for correct length
        def test_routine(start, end=None, length=None):
            random_walk_ts = random_walk_timeseries(start=start, end=end, length=length)
            self.assertEqual(len(random_walk_ts), length_assert)

        for length_assert in [1, 2, 5, 10, 100]:
            test_routine(start=0, length=length_assert)
            test_routine(start=0, end=length_assert - 1)
            test_routine(start=pd.Timestamp("2000-01-01"), length=length_assert)
            end_date = generate_index(
                start=pd.Timestamp("2000-01-01"), length=length_assert
            )[-1]
            test_routine(start=pd.Timestamp("2000-01-01"), end=end_date)

    def test_holidays_timeseries(self):
        time_index_1 = pd.date_range(
            periods=365 * 3, freq="D", start=pd.Timestamp("2012-01-01")
        )
        time_index_2 = pd.date_range(
            periods=365 * 3, freq="D", start=pd.Timestamp("2014-12-24")
        )
        time_index_3 = pd.date_range(
            periods=10, freq="Y", start=pd.Timestamp("1950-01-01")
        ) + pd.Timedelta(days=1)

        # testing we have at least one holiday flag in each year
        def test_routine(
            time_index,
            country_code,
            until: Union[int, pd.Timestamp, str] = 0,
            add_length=0,
        ):
            ts = holidays_timeseries(
                time_index, country_code, until=until, add_length=add_length
            )
            self.assertTrue(
                all(ts.pd_dataframe().groupby(pd.Grouper(freq="y")).sum().values)
            )

        for time_index in [time_index_1, time_index_2, time_index_3]:
            for country_code in ["US", "CH", "AR"]:
                test_routine(time_index, country_code)

        # test extend time index
        test_routine(time_index_1, "US", add_length=365)
        test_routine(time_index_1, "CH", until="2016-01-01")
        test_routine(time_index_1, "CH", until="20160101")
        test_routine(time_index_1, "AR", until=pd.Timestamp("2016-01-01"))

        # test overflow
        with self.assertRaises(ValueError):
            holidays_timeseries(time_index_1, "US", add_length=99999)

        # test date is too short
        with self.assertRaises(ValueError):
            holidays_timeseries(time_index_2, "US", until="2016-01-01")

        # test wrong timestamp
        with self.assertRaises(ValueError):
            holidays_timeseries(time_index_3, "US", until=163)

    def test_generate_index(self):
        def test_routine(
            expected_length,
            expected_start,
            expected_end,
            start,
            end=None,
            length=None,
            freq=None,
        ):
            index = generate_index(start=start, end=end, length=length, freq=freq)
            self.assertEqual(len(index), expected_length)
            self.assertEqual(index[0], expected_start)
            self.assertEqual(index[-1], expected_end)

        for length in [1, 2, 5, 50]:
            for start in [0, 1, 9]:

                # test pd.RangeIndex with varying step sizes
                for step in [1, 2, 4]:
                    expected_start = start
                    expected_end = start + (length - 1) * step
                    freq = None if step == 1 else step
                    test_routine(
                        expected_length=length,
                        expected_start=expected_start,
                        expected_end=expected_end,
                        start=start,
                        length=length,
                        freq=freq,
                    )

                    test_routine(
                        expected_length=length,
                        expected_start=expected_start,
                        expected_end=expected_end,
                        start=start,
                        end=expected_end,
                        freq=step,
                    )

                    test_routine(
                        expected_length=length,
                        expected_start=expected_start,
                        expected_end=expected_end,
                        start=None,
                        end=expected_end,
                        length=length,
                        freq=step,
                    )

                    if start == 0:
                        continue

                    # test pd.DatetimeIndex with a start date within 01 and 09
                    start_date = pd.Timestamp(f"2000-01-0{start}")
                    dates = generate_index(
                        start=start_date,
                        length=length,
                        freq="D" if step == 1 else f"{step}D",
                    )
                    start_assert, end_assert = dates[0], dates[-1]
                    test_routine(
                        expected_length=length,
                        expected_start=start_assert,
                        expected_end=end_assert,
                        start=start_assert,
                        length=length,
                        freq="D" if step == 1 else f"{step}D",
                    )
                    test_routine(
                        expected_length=length,
                        expected_start=start_assert,
                        expected_end=end_assert,
                        start=start_assert,
                        end=end_assert,
                        freq="D" if step == 1 else f"{step}D",
                    )
                    test_routine(
                        expected_length=length,
                        expected_start=start_assert,
                        expected_end=end_assert,
                        start=None,
                        end=end_assert,
                        length=length,
                        freq="D" if step == 1 else f"{step}D",
                    )

        # `start`, `end` and `length` cannot both be set simultaneously
        with self.assertRaises(ValueError):
            generate_index(start=0, end=9, length=10)
        # same as above but `start` defaults to timestamp '2000-01-01' in all timeseries generation functions
        with self.assertRaises(ValueError):
            linear_timeseries(end=9, length=10)

        # exactly two of [`start`, `end`, `length`] must be set
        with self.assertRaises(ValueError):
            generate_index(start=0)
        with self.assertRaises(ValueError):
            generate_index(start=None, end=1)
        with self.assertRaises(ValueError):
            generate_index(start=None, end=None, length=10)

        # `start` and `end` must have same type
        with self.assertRaises(ValueError):
            generate_index(start=0, end=pd.Timestamp("2000-01-01"))
        with self.assertRaises(ValueError):
            generate_index(start=pd.Timestamp("2000-01-01"), end=10)

    def test_autoregressive_timeseries(self):
        # testing for correct length
        def test_length(start, end=None, length=None):
            autoregressive_ts = autoregressive_timeseries(
                coef=[-1, 1.618], start=start, end=end, length=length
            )
            self.assertEqual(len(autoregressive_ts), length_assert)

        # testing for correct calculation
        def test_calculation(coef):
            autoregressive_values = autoregressive_timeseries(
                coef=coef, length=100
            ).values()
            for idx, val in enumerate(autoregressive_values[len(coef) :]):
                self.assertTrue(
                    val
                    == np.dot(
                        coef, autoregressive_values[idx : idx + len(coef)].ravel()
                    )
                )

        for length_assert in [1, 2, 5, 10, 100]:
            test_length(start=0, length=length_assert)
            test_length(start=0, end=length_assert - 1)
            test_length(start=pd.Timestamp("2000-01-01"), length=length_assert)
            end_date = generate_index(
                start=pd.Timestamp("2000-01-01"), length=length_assert
            )[-1]
            test_length(start=pd.Timestamp("2000-01-01"), end=end_date)

        for coef_assert in [[-1], [-1, 1.618], [1, 2, 3], list(range(10))]:
            test_calculation(coef=coef_assert)



import unittest

import numpy as np
import pandas as pd
import pytest

from darts import TimeSeries
from darts.dataprocessing.transformers import Scaler
from darts.datasets import AirPassengersDataset
from darts.logging import get_logger
from darts.models import (
    ARIMA,
    AutoARIMA,
    CatBoostModel,
    LightGBMModel,
    LinearRegressionModel,
    NaiveSeasonal,
)
from darts.tests.base_test_class import DartsBaseTestClass
from darts.utils import timeseries_generation as tg

try:
    import torch

    from darts.models import (
        BlockRNNModel,
        NBEATSModel,
        NLinearModel,
        RNNModel,
        TCNModel,
        TFTModel,
        TransformerModel,
    )
    from darts.utils.likelihood_models import GaussianLikelihood

    TORCH_AVAILABLE = True
except ImportError:
    logger = get_logger(__name__)
    logger.warning(
        "Torch not installed - will be skipping historical forecasts tests for torch models"
    )
    TORCH_AVAILABLE = False

models_reg_no_cov_cls_kwargs = [
    (LinearRegressionModel, {"lags": 8}, (8, 1)),
    (CatBoostModel, {"lags": 6}, (6, 1)),
    (LightGBMModel, {"lags": 4}, (4, 1)),
]

models_reg_cov_cls_kwargs = [
    # target + past covariates
    (LinearRegressionModel, {"lags": 4, "lags_past_covariates": 6}, (6, 1)),
    # target + past covariates + outputchunk > 3, 6 > 3
    (
        LinearRegressionModel,
        {"lags": 3, "lags_past_covariates": 6, "output_chunk_length": 5},
        (6, 5),
    ),
    # target + future covariates, 2 because to predict x, require x and x+1
    (LinearRegressionModel, {"lags": 4, "lags_future_covariates": [0, 1]}, (4, 2)),
    # target + fut cov + output_chunk_length > 3,
    (
        LinearRegressionModel,
        {"lags": 2, "lags_future_covariates": [1, 2], "output_chunk_length": 5},
        (2, 5),
    ),
    # fut cov + output_chunk_length > 3, 5 > 2
    (
        LinearRegressionModel,
        {"lags_future_covariates": [0, 1], "output_chunk_length": 5},
        (0, 5),
    ),
    # past cov only
    (LinearRegressionModel, {"lags_past_covariates": 6}, (6, 1)),
    # fut cov only
    (LinearRegressionModel, {"lags_future_covariates": [0, 1]}, (0, 2)),
    # fut + past cov only
    (
        LinearRegressionModel,
        {"lags_past_covariates": 6, "lags_future_covariates": [0, 1]},
        (6, 2),
    ),
    # all
    (
        LinearRegressionModel,
        {"lags": 3, "lags_past_covariates": 6, "lags_future_covariates": [0, 1]},
        (6, 2),
    ),
]

if TORCH_AVAILABLE:
    IN_LEN = 24
    OUT_LEN = 12

    NB_EPOCH = 1

    models_torch_cls_kwargs = [
        (
            BlockRNNModel,
            {
                "input_chunk_length": IN_LEN,
                "output_chunk_length": OUT_LEN,
                "model": "RNN",
                "hidden_dim": 10,
                "n_rnn_layers": 1,
                "batch_size": 32,
                "n_epochs": NB_EPOCH,
            },
            # Min of lags needed and max of lags needed
            (IN_LEN, OUT_LEN),
            "PastCovariates",
        ),
        (
            RNNModel,
            {
                "input_chunk_length": IN_LEN,
                "model": "RNN",
                "hidden_dim": 10,
                "batch_size": 32,
                "n_epochs": NB_EPOCH,
            },
            # autoregressive model
            (IN_LEN, 1),
            "DualCovariates",
        ),
        (
            RNNModel,
            {
                "input_chunk_length": IN_LEN,
                "training_length": 12,
                "n_epochs": NB_EPOCH,
                "likelihood": GaussianLikelihood(),
            },
            (IN_LEN, 1),
            "DualCovariates",
        ),
        (
            TCNModel,
            {
                "input_chunk_length": IN_LEN,
                "output_chunk_length": OUT_LEN,
                "n_epochs": NB_EPOCH,
                "batch_size": 32,
            },
            (IN_LEN, OUT_LEN),
            "PastCovariates",
        ),
        (
            TransformerModel,
            {
                "input_chunk_length": IN_LEN,
                "output_chunk_length": OUT_LEN,
                "d_model": 16,
                "nhead": 2,
                "num_encoder_layers": 2,
                "num_decoder_layers": 2,
                "dim_feedforward": 16,
                "batch_size": 32,
                "n_epochs": NB_EPOCH,
            },
            (IN_LEN, OUT_LEN),
            "PastCovariates",
        ),
        (
            NBEATSModel,
            {
                "input_chunk_length": IN_LEN,
                "output_chunk_length": OUT_LEN,
                "num_stacks": 4,
                "num_blocks": 1,
                "num_layers": 2,
                "layer_widths": 12,
                "n_epochs": NB_EPOCH,
            },
            (IN_LEN, OUT_LEN),
            "PastCovariates",
        ),
        (
            TFTModel,
            {
                "input_chunk_length": IN_LEN,
                "output_chunk_length": OUT_LEN,
                "hidden_size": 16,
                "lstm_layers": 1,
                "num_attention_heads": 4,
                "add_relative_index": True,
                "n_epochs": NB_EPOCH,
            },
            (IN_LEN, OUT_LEN),
            "MixedCovariates",
        ),
        (
            NLinearModel,
            {
                "input_chunk_length": IN_LEN,
                "output_chunk_length": OUT_LEN,
                "n_epochs": NB_EPOCH,
            },
            (IN_LEN, OUT_LEN),
            "MixedCovariates",
        ),
    ]


class HistoricalforecastTestCase(DartsBaseTestClass):
    np.random.seed(42)
    if TORCH_AVAILABLE:
        torch.manual_seed(42)

    # real timeseries for functionality tests
    ts_val_length = 72
    ts_passengers = AirPassengersDataset().load()
    scaler = Scaler()
    ts_passengers = scaler.fit_transform(ts_passengers)
    ts_pass_train, ts_pass_val = (
        ts_passengers[:-ts_val_length],
        ts_passengers[-ts_val_length:],
    )

    # an additional noisy series
    ts_pass_train_1 = ts_pass_train + 0.01 * tg.gaussian_timeseries(
        length=len(ts_pass_train),
        freq=ts_pass_train.freq_str,
        start=ts_pass_train.start_time(),
    )

    ts_past_cov_train = tg.gaussian_timeseries(
        length=len(ts_pass_train),
        freq=ts_pass_train.freq_str,
        start=ts_pass_train.start_time(),
    )

    ts_fut_cov_train = tg.gaussian_timeseries(
        length=len(ts_pass_train),
        freq=ts_pass_train.freq_str,
        start=ts_pass_train.start_time(),
    )

    ts_past_cov_valid_same_start = tg.gaussian_timeseries(
        length=len(ts_pass_val),
        freq=ts_pass_val.freq_str,
        start=ts_pass_val.start_time(),
    )

    ts_past_cov_valid_10_bef_start = tg.gaussian_timeseries(
        length=len(ts_pass_val) + 10,
        freq=ts_pass_val.freq_str,
        start=ts_pass_val.start_time() - 10 * ts_pass_val.freq,
    )
    ts_past_cov_valid_5_aft_start = tg.gaussian_timeseries(
        length=len(ts_pass_val) - 5,
        freq=ts_pass_val.freq_str,
        start=ts_pass_val.start_time() + 5 * ts_pass_val.freq,
    )

    ts_fut_cov_valid_same_start = tg.gaussian_timeseries(
        length=len(ts_pass_val),
        freq=ts_pass_val.freq_str,
        start=ts_pass_val.start_time(),
    )

    ts_fut_cov_valid_16_bef_start = tg.gaussian_timeseries(
        length=len(ts_pass_val) + 16,
        freq=ts_pass_val.freq_str,
        start=ts_pass_val.start_time() - 16 * ts_pass_val.freq,
    )
    ts_fut_cov_valid_7_aft_start = tg.gaussian_timeseries(
        length=len(ts_pass_val) - 7,
        freq=ts_pass_val.freq_str,
        start=ts_pass_val.start_time() + 7 * ts_pass_val.freq,
    )

    # RangeIndex timeseries
    ts_passengers_range = TimeSeries.from_values(ts_passengers.values())
    ts_pass_train_range, ts_pass_val_range = (
        ts_passengers_range[:-ts_val_length],
        ts_passengers_range[-ts_val_length:],
    )

    ts_past_cov_train_range = tg.gaussian_timeseries(
        length=len(ts_pass_train_range),
        freq=ts_pass_train_range.freq_str,
        start=ts_pass_train_range.start_time(),
    )

    # same starting point
    ts_past_cov_valid_range_same_start = tg.gaussian_timeseries(
        length=len(ts_pass_val_range),
        freq=ts_pass_val_range.freq_str,
        start=ts_pass_val_range.start_time(),
    )

    def test_historical_forecasts_transferrable_future_cov_local_models(self):
        model = ARIMA()
        assert model.min_train_series_length == 30
        series = tg.sine_timeseries(length=31)
        res = model.historical_forecasts(
            series, future_covariates=series, retrain=True, forecast_horizon=1
        )
        # ARIMA has a minimum train length of 30, with horizon=1, we expect one forecast at last point
        # (series has length 31)
        assert len(res) == 1
        assert series.end_time() == res.time_index[0]

        model.fit(series, future_covariates=series)
        res = model.historical_forecasts(
            series, future_covariates=series, retrain=False, forecast_horizon=1
        )
        # currently even though transferrable local models would allow , the models currently still take the
        # min_train_length as input for historical forecast predictions (due to extreme_lags not differentiating
        # between fit and predict)
        # (series has length 31)
        assert len(res) == 1
        assert series.end_time() == res.time_index[0]

    def test_historical_forecasts_future_cov_local_models(self):
        model = AutoARIMA()
        assert model.min_train_series_length == 10
        series = tg.sine_timeseries(length=11)
        res = model.historical_forecasts(
            series, future_covariates=series, retrain=True, forecast_horizon=1
        )
        # AutoARIMA has a minimum train length of 10, with horizon=1, we expect one forecast at last point
        # (series has length 11)
        assert len(res) == 1
        assert series.end_time() == res.time_index[0]

        model.fit(series, future_covariates=series)
        with pytest.raises(ValueError) as msg:
            model.historical_forecasts(
                series, future_covariates=series, retrain=False, forecast_horizon=1
            )
        assert str(msg.value).startswith(
            "FutureCovariatesLocalForecastingModel does not support historical forecasting "
            "with `retrain` set to `False`"
        )

    def test_historical_forecasts_local_models(self):
        model = NaiveSeasonal()
        assert model.min_train_series_length == 3
        series = tg.sine_timeseries(length=4)
        res = model.historical_forecasts(series, retrain=True, forecast_horizon=1)
        # NaiveSeasonal has a minimum train length of 3, with horizon=1, we expect one forecast at last point
        # (series has length 4)
        assert len(res) == 1
        assert series.end_time() == res.time_index[0]

        model.fit(series)
        with pytest.raises(ValueError) as msg:
            model.historical_forecasts(series, retrain=False, forecast_horizon=1)
        assert str(msg.value).startswith(
            "LocalForecastingModel does not support historical forecasting with `retrain` set to `False`"
        )

    def test_historical_forecasts(self):
        train_length = 10
        forecast_horizon = 8
        # if no fit and retrain=false, should fit at fist iteration
        for model_cls, kwargs, bounds in models_reg_no_cov_cls_kwargs:
            model = model_cls(
                **kwargs,
            )

            # time index
            forecasts = model.historical_forecasts(
                series=self.ts_pass_val,
                forecast_horizon=forecast_horizon,
                stride=1,
                train_length=train_length,
                retrain=True,
                overlap_end=False,
            )

            theorical_forecast_length = (
                self.ts_val_length
                - max(
                    [
                        (
                            bounds[0] + bounds[1] + 1
                        ),  # +1 as sklearn models require min 2 train samples
                        train_length,
                    ]
                )  # because we train
                - forecast_horizon  # because we have overlap_end = False
                + 1  # because we include the first element
            )

            self.assertTrue(
                len(forecasts) == theorical_forecast_length,
                f"Model {model_cls.__name__} does not return the right number of historical forecasts in the case "
                f"of retrain=True and overlap_end=False, and a time index of type DateTimeIndex. "
                f"Expected {theorical_forecast_length}, got {len(forecasts)}",
            )

            # range index
            forecasts = model.historical_forecasts(
                series=self.ts_pass_val_range,
                forecast_horizon=forecast_horizon,
                train_length=train_length,
                stride=1,
                retrain=True,
                overlap_end=False,
            )

            self.assertTrue(
                len(forecasts) == theorical_forecast_length,
                f"Model {model_cls.__name__} does not return the right number of historical forecasts in the case "
                f"of retrain=True, overlap_end=False, and a time index of type RangeIndex."
                f"Expected {theorical_forecast_length}, got {len(forecasts)}",
            )

            # stride 2
            forecasts = model.historical_forecasts(
                series=self.ts_pass_val_range,
                forecast_horizon=forecast_horizon,
                train_length=train_length,
                stride=2,
                retrain=True,
                overlap_end=False,
            )

            theorical_forecast_length = np.floor(
                (
                    (
                        self.ts_val_length
                        - max(
                            [
                                (
                                    bounds[0] + bounds[1] + 1
                                ),  # +1 as sklearn models require min 2 train samples
                                train_length,
                            ]
                        )  # because we train
                        - forecast_horizon  # because we have overlap_end = False
                        + 1  # because we include the first element
                    )
                    - 1
                )
                / 2
                + 1  # because of stride
            )  # if odd number of elements, we keep the floor

            self.assertTrue(
                len(forecasts) == theorical_forecast_length,
                f"Model {model_cls.__name__} does not return the right number of historical forecasts in the case "
                f"of retrain=True and overlap_end=False and stride=2. "
                f"Expected {theorical_forecast_length}, got {len(forecasts)}",
            )

            # stride 3
            forecasts = model.historical_forecasts(
                series=self.ts_pass_val_range,
                forecast_horizon=forecast_horizon,
                train_length=train_length,
                stride=3,
                retrain=True,
                overlap_end=False,
            )

            theorical_forecast_length = np.floor(
                (
                    (
                        self.ts_val_length
                        - max(
                            [
                                (
                                    bounds[0] + bounds[1] + 1
                                ),  # +1 as sklearn models require min 2 train samples
                                train_length,
                            ]
                        )  # because we train
                        - forecast_horizon  # because we have overlap_end = False
                        + 1  # because we include the first element
                    )
                    - 1
                )  # the first is always included, so we calculate a modulo on the rest
                / 3  # because of stride
                + 1  # and we readd the first
            )  # if odd number of elements, we keep the floor

            # Here to adapt if forecast_horizon or train_length change
            self.assertTrue(
                len(forecasts) == theorical_forecast_length,
                f"Model {model_cls.__name__} does not return the right number of historical forecasts in the case "
                f"of retrain=True and overlap_end=False and stride=3. "
                f"Expected {theorical_forecast_length}, got {len(forecasts)}",
            )

            # last points only False
            forecasts = model.historical_forecasts(
                series=self.ts_pass_val_range,
                forecast_horizon=forecast_horizon,
                train_length=train_length,
                stride=1,
                retrain=True,
                overlap_end=False,
                last_points_only=False,
            )

            theorical_forecast_length = (
                self.ts_val_length
                - max(
                    [
                        (
                            bounds[0] + bounds[1] + 1
                        ),  # +1 as sklearn models require min 2 train samples
                        train_length,
                    ]
                )  # because we train
                - forecast_horizon  # because we have overlap_end = False
                + 1  # because we include the first element
            )

            self.assertTrue(
                len(forecasts) == theorical_forecast_length,
                f"Model {model_cls} does not return the right number of historical forecasts in the case of "
                f"retrain=True and overlap_end=False, and last_points_only=False. "
                f"expected {theorical_forecast_length}, got {len(forecasts)}",
            )

            self.assertTrue(
                len(forecasts[0]) == forecast_horizon,
                f"Model {model_cls} does not return forecast_horizon points per historical forecast in the case of "
                f"retrain=True and overlap_end=False, and last_points_only=False",
            )

    def test_sanity_check_invalid_start(self):
        timeidx_ = tg.linear_timeseries(length=10)
        rangeidx_step1 = tg.linear_timeseries(start=0, length=10, freq=1)
        rangeidx_step2 = tg.linear_timeseries(start=0, length=10, freq=2)

        # index too large
        with pytest.raises(ValueError) as msg:
            LinearRegressionModel(lags=1).historical_forecasts(timeidx_, start=11)
        assert str(msg.value).startswith("`start` index `11` is out of bounds")
        with pytest.raises(ValueError) as msg:
            LinearRegressionModel(lags=1).historical_forecasts(rangeidx_step1, start=11)
        assert str(msg.value).startswith("`start` index `11` is out of bounds")
        with pytest.raises(ValueError) as msg:
            LinearRegressionModel(lags=1).historical_forecasts(rangeidx_step2, start=11)
        assert str(msg.value).startswith("The provided point is not a valid index")

        # value too low
        with pytest.raises(ValueError) as msg:
            LinearRegressionModel(lags=1).historical_forecasts(
                timeidx_, start=timeidx_.start_time() - timeidx_.freq
            )
        assert str(msg.value).startswith(
            "`start` time `1999-12-31 00:00:00` is before the first timestamp `2000-01-01 00:00:00`"
        )
        with pytest.raises(ValueError) as msg:
            LinearRegressionModel(lags=1).historical_forecasts(
                rangeidx_step1, start=rangeidx_step1.start_time() - rangeidx_step1.freq
            )
        assert str(msg.value).startswith("if `start` is an integer, must be `>= 0`")
        with pytest.raises(ValueError) as msg:
            LinearRegressionModel(lags=1).historical_forecasts(
                rangeidx_step2, start=rangeidx_step2.start_time() - rangeidx_step2.freq
            )
        assert str(msg.value).startswith("if `start` is an integer, must be `>= 0`")

        # value too high
        with pytest.raises(ValueError) as msg:
            LinearRegressionModel(lags=1).historical_forecasts(
                timeidx_, start=timeidx_.end_time() + timeidx_.freq
            )
        assert str(msg.value).startswith(
            "`start` time `2000-01-11 00:00:00` is after the last timestamp `2000-01-10 00:00:00`"
        )
        with pytest.raises(ValueError) as msg:
            LinearRegressionModel(lags=1).historical_forecasts(
                rangeidx_step1, start=rangeidx_step1.end_time() + rangeidx_step1.freq
            )
        assert str(msg.value).startswith("`start` index `10` is out of bounds")
        with pytest.raises(ValueError) as msg:
            LinearRegressionModel(lags=1).historical_forecasts(
                rangeidx_step2, start=rangeidx_step2.end_time() + rangeidx_step2.freq
            )
        assert str(msg.value).startswith(
            "`start` index `20` is larger than the last index `18`"
        )

    def test_regression_auto_start_multiple_no_cov(self):
        train_length = 15
        forecast_horizon = 10
        for model_cls, kwargs, bounds in models_reg_no_cov_cls_kwargs:
            model = model_cls(
                **kwargs,
            )
            model.fit(self.ts_pass_train)

            forecasts = model.historical_forecasts(
                series=[self.ts_pass_val, self.ts_pass_val],
                forecast_horizon=forecast_horizon,
                train_length=train_length,
                stride=1,
                retrain=True,
                overlap_end=False,
            )

            self.assertTrue(
                len(forecasts) == 2,
                f"Model {model_cls} did not return a list of historical forecasts",
            )

            theorical_forecast_length = (
                self.ts_val_length
                - max(
                    [
                        (
                            bounds[0] + bounds[1] + 1
                        ),  # +1 as sklearn models require min 2 train samples
                        train_length,
                    ]
                )  # because we train
                - forecast_horizon  # because we have overlap_end = False
                + 1  # because we include the first element
            )

            self.assertTrue(
                len(forecasts[0]) == len(forecasts[1]) == theorical_forecast_length,
                f"Model {model_cls.__name__} does not return the right number of historical forecasts in the case "
                f"of retrain=True and overlap_end=False, and a time index of type DateTimeIndex. "
                f"Expected {theorical_forecast_length}, got {len(forecasts[0])} and {len(forecasts[1])}",
            )

    @pytest.mark.slow
    @unittest.skipUnless(
        TORCH_AVAILABLE,
        "Torch not available. auto start and multiple time series for torch models will be skipped.",
    )
    def test_torch_auto_start_multiple_no_cov(self):
        forecast_hrz = 10
        for model_cls, kwargs, bounds, _ in models_torch_cls_kwargs:
            model = model_cls(
                random_state=0,
                **kwargs,
            )
            model.fit(self.ts_pass_train)

            # check historical forecasts for several time series,
            # retrain True and overlap_end False
            forecasts = model.historical_forecasts(
                series=[self.ts_pass_val, self.ts_pass_val],
                forecast_horizon=forecast_hrz,
                stride=1,
                retrain=True,
                overlap_end=False,
            )
            self.assertTrue(
                len(forecasts) == 2,
                f"Model {model_cls} did not return a list of historical forecasts",
            )
            # If retrain=True and overlap_end=False, as ts has 72 values, we can only forecast
            # (target length)-(training length=input_chunk_length+output_chunk_length) - (horizon - 1)
            # indeed we start to predict after the first trainable point (input_chunk_length+output_chunk_length)
            # and we stop in this case (overlap_end=False) at the end_time:
            # target.end_time() - (horizon - 1) * target.freq

            # explanation:
            # (bounds): train sample length
            # (horizon - 1): with overlap_end=False, if entire horizon is available (overlap_end=False),
            # we can predict 1
            theorical_forecast_length = (
                self.ts_val_length - (bounds[0] + bounds[1]) - (forecast_hrz - 1)
            )
            self.assertTrue(
                len(forecasts[0]) == len(forecasts[1]) == theorical_forecast_length,
                f"Model {model_cls} does not return the right number of historical forecasts in the case of "
                f"retrain=True and overlap_end=False. "
                f"Expected {theorical_forecast_length}, got {len(forecasts[0])} and {len(forecasts[1])}",
            )

            model = model_cls(
                random_state=0,
                **kwargs,
            )

            model.fit(self.ts_pass_train)
            # check historical forecasts for several time series,
            # retrain True and overlap_end True
            forecasts = model.historical_forecasts(
                series=[self.ts_pass_val, self.ts_pass_val],
                forecast_horizon=forecast_hrz,
                stride=1,
                retrain=True,
                overlap_end=True,
            )

            self.assertTrue(
                len(forecasts) == 2,
                f"Model {model_cls} did not return a list of historical forecasts",
            )
            theorical_forecast_length = (
                self.ts_val_length
                - (bounds[0] + bounds[1])  # train sample length
                - 0  # with overlap_end=True, we are not restricted by the end of the series or horizon
            )
            self.assertTrue(
                len(forecasts[0]) == len(forecasts[1]) == theorical_forecast_length
            )

            model = model_cls(
                random_state=0,
                **kwargs,
            )
            model.fit(self.ts_pass_train)
            # check historical forecasts for several time series,
            # retrain False and overlap_end False
            forecasts = model.historical_forecasts(
                series=[self.ts_pass_val, self.ts_pass_val],
                forecast_horizon=forecast_hrz,
                stride=1,
                retrain=False,
                overlap_end=False,
            )

            self.assertTrue(
                len(forecasts) == 2,
                f"Model {model_cls} did not return a list of historical forecasts",
            )
            theorical_forecast_length = (
                self.ts_val_length
                - bounds[0]  # prediction input sample length
                - (
                    forecast_hrz - 1
                )  # overlap_end=False -> if entire horizon is available, we can predict 1
            )
            self.assertTrue(
                len(forecasts[0]) == len(forecasts[1]) == theorical_forecast_length
            )

            model = model_cls(
                random_state=0,
                **kwargs,
            )
            model.fit(self.ts_pass_train)
            # check historical forecasts for several time series,
            # retrain False and overlap_end True
            forecasts = model.historical_forecasts(
                series=[self.ts_pass_val, self.ts_pass_val],
                forecast_horizon=forecast_hrz,
                stride=1,
                retrain=False,
                overlap_end=True,
            )

            self.assertTrue(
                len(forecasts) == 2,
                f"Model {model_cls} did not return a list of historical forecasts",
            )
            theorical_forecast_length = (
                self.ts_val_length
                - bounds[0]  # prediction input sample length
                - 0  # overlap_end=False -> we are not restricted by the end of the series or horizon
            )
            self.assertTrue(
                len(forecasts[0]) == len(forecasts[1]) == theorical_forecast_length
            )

    def test_regression_auto_start_multiple_with_cov_retrain(self):
        forecast_hrz = 10
        for model_cls, kwargs, bounds in models_reg_cov_cls_kwargs:
            model = model_cls(
                random_state=0,
                **kwargs,
            )

            forecasts_retrain = model.historical_forecasts(
                series=[self.ts_pass_val, self.ts_pass_val],
                past_covariates=[
                    self.ts_past_cov_valid_same_start,
                    self.ts_past_cov_valid_same_start,
                ]
                if "lags_past_covariates" in kwargs
                else None,
                future_covariates=[
                    self.ts_past_cov_valid_same_start,
                    self.ts_past_cov_valid_same_start,
                ]
                if "lags_future_covariates" in kwargs
                else None,
                last_points_only=True,
                forecast_horizon=forecast_hrz,
                stride=1,
                retrain=True,
                overlap_end=False,
            )

            self.assertTrue(
                len(forecasts_retrain) == 2,
                f"Model {model_cls} did not return a list of historical forecasts",
            )

            (
                min_target_lag,
                max_target_lag,
                min_past_cov_lag,
                max_past_cov_lag,
                min_future_cov_lag,
                max_future_cov_lag,
            ) = model.extreme_lags

            past_lag = min(
                min_target_lag if min_target_lag else 0,
                min_past_cov_lag if min_past_cov_lag else 0,
                min_future_cov_lag
                if min_future_cov_lag is not None and min_future_cov_lag < 0
                else 0,
            )

            future_lag = (
                max_future_cov_lag
                if max_future_cov_lag is not None and max_future_cov_lag > 0
                else 0
            )
            # length input - biggest past lag - biggest future lag - forecast horizon - output_chunk_length
            theorical_retrain_forecast_length = len(self.ts_pass_val) - (
                -past_lag
                + future_lag
                + forecast_hrz
                + kwargs.get("output_chunk_length", 1)
            )

            self.assertTrue(
                len(forecasts_retrain[0])
                == len(forecasts_retrain[1])
                == theorical_retrain_forecast_length,
                f"Model {model_cls} does not return the right number of historical forecasts in the case of "
                f"retrain=True and overlap_end=False. "
                f"Expected {theorical_retrain_forecast_length}, got {len(forecasts_retrain[0])} "
                f"and {len(forecasts_retrain[1])}",
            )

            # with last_points_only=True: start is shifted by biggest past lag + training timestamps
            # (forecast horizon + output_chunk_length)
            expected_start = (
                self.ts_pass_val.start_time()
                + (-past_lag + forecast_hrz + kwargs.get("output_chunk_length", 1))
                * self.ts_pass_val.freq
            )
            self.assertEqual(forecasts_retrain[0].start_time(), expected_start)

            # end is shifted back by the biggest future lag
            expected_end = (
                self.ts_pass_val.end_time() - future_lag * self.ts_pass_val.freq
            )
            self.assertEqual(forecasts_retrain[0].end_time(), expected_end)

    def test_regression_auto_start_multiple_with_cov_no_retrain(self):
        forecast_hrz = 10
        for model_cls, kwargs, bounds in models_reg_cov_cls_kwargs:
            model = model_cls(
                random_state=0,
                **kwargs,
            )

            model.fit(
                series=[self.ts_pass_val, self.ts_pass_val],
                past_covariates=[
                    self.ts_past_cov_valid_same_start,
                    self.ts_past_cov_valid_same_start,
                ]
                if "lags_past_covariates" in kwargs
                else None,
                future_covariates=[
                    self.ts_past_cov_valid_same_start,
                    self.ts_past_cov_valid_same_start,
                ]
                if "lags_future_covariates" in kwargs
                else None,
            )
            forecasts_no_retrain = model.historical_forecasts(
                series=[self.ts_pass_val, self.ts_pass_val],
                past_covariates=[
                    self.ts_past_cov_valid_same_start,
                    self.ts_past_cov_valid_same_start,
                ]
                if "lags_past_covariates" in kwargs
                else None,
                future_covariates=[
                    self.ts_past_cov_valid_same_start,
                    self.ts_past_cov_valid_same_start,
                ]
                if "lags_future_covariates" in kwargs
                else None,
                last_points_only=True,
                forecast_horizon=forecast_hrz,
                stride=1,
                retrain=False,
                overlap_end=False,
            )

            (
                min_target_lag,
                max_target_lag,
                min_past_cov_lag,
                max_past_cov_lag,
                min_future_cov_lag,
                max_future_cov_lag,
            ) = model.extreme_lags

            past_lag = min(
                min_target_lag if min_target_lag else 0,
                min_past_cov_lag if min_past_cov_lag else 0,
                min_future_cov_lag if min_future_cov_lag else 0,
            )

            future_lag = (
                max_future_cov_lag
                if max_future_cov_lag is not None and max_future_cov_lag > 0
                else 0
            )

            # with last_points_only=True: start is shifted by the biggest past lag plus the forecast horizon
            expected_start = (
                self.ts_pass_val.start_time()
                + (-past_lag + forecast_hrz - 1) * self.ts_pass_val.freq
            )
            self.assertEqual(forecasts_no_retrain[0].start_time(), expected_start)

            # end is shifted by the biggest future lag
            expected_end = (
                self.ts_pass_val.end_time() - future_lag * self.ts_pass_val.freq
            )
            self.assertEqual(forecasts_no_retrain[0].end_time(), expected_end)

    @pytest.mark.slow
    @unittest.skipUnless(
        TORCH_AVAILABLE,
        "Torch not available. auto start and multiple time series for torch models and covariates "
        "will be skipped.",
    )
    def test_torch_auto_start_with_cov(self):
        forecast_hrz = 10
        # Past covariates only
        for model_cls, kwargs, bounds, type in models_torch_cls_kwargs:

            if type == "DualCovariates":
                continue

            model = model_cls(
                random_state=0,
                **kwargs,
            )
            model.fit(self.ts_pass_train, self.ts_past_cov_train)

            # same start
            forecasts = model.historical_forecasts(
                series=[self.ts_pass_val, self.ts_pass_val],
                past_covariates=[
                    self.ts_past_cov_valid_same_start,
                    self.ts_past_cov_valid_same_start,
                ],
                forecast_horizon=forecast_hrz,
                stride=1,
                retrain=True,
                overlap_end=False,
            )

            self.assertTrue(
                len(forecasts) == 2,
                f"Model {model_cls} did not return a list of historical forecasts",
            )

            theorical_forecast_length = (
                self.ts_val_length
                - (bounds[0] + bounds[1])  # train sample length
                - (forecast_hrz - 1)  # if entire horizon is available, we can predict 1
                - 0  # past covs have same start as target -> no shift
                - 0  # we don't have future covs in output chunk -> no shift
            )
            self.assertTrue(
                len(forecasts[0]) == len(forecasts[1]) == theorical_forecast_length,
                f"Model {model_cls} does not return the right number of historical forecasts in case "
                f"of retrain=True and overlap_end=False and past_covariates with same start. "
                f"Expected {theorical_forecast_length}, got {len(forecasts[0])} and {len(forecasts[1])}",
            )

            model = model_cls(
                random_state=0,
                **kwargs,
            )
            model.fit(self.ts_pass_train, past_covariates=self.ts_past_cov_train)

            # start before, after
            forecasts = model.historical_forecasts(
                series=[self.ts_pass_val, self.ts_pass_val],
                past_covariates=[
                    self.ts_past_cov_valid_5_aft_start,
                    self.ts_past_cov_valid_10_bef_start,
                ],
                forecast_horizon=forecast_hrz,
                stride=1,
                retrain=True,
                overlap_end=False,
            )
            theorical_forecast_length = (
                self.ts_val_length
                - (bounds[0] + bounds[1])  # train sample length
                - (forecast_hrz - 1)  # if entire horizon is available, we can predict 1
                - 5  # past covs start 5 later -> shift
                - 0  # we don't have future covs in output chunk -> no shift
            )
            self.assertTrue(
                len(forecasts[0]) == theorical_forecast_length,
                f"Model {model_cls} does not return the right number of historical forecasts in case "
                f"of retrain=True and overlap_end=False and past_covariates starting after. "
                f"Expected {theorical_forecast_length}, got {len(forecasts[0])}",
            )
            theorical_forecast_length = (
                self.ts_val_length
                - (bounds[0] + bounds[1])  # train sample length
                - (forecast_hrz - 1)  # if entire horizon is available, we can predict 1
                - 0  # past covs have same start as target -> no shift
                - 0  # we don't have future covs in output chunk -> no shift
            )
            self.assertTrue(
                len(forecasts[1]) == theorical_forecast_length,
                f"Model {model_cls} does not return the right number of historical forecasts in case "
                f"of retrain=True and overlap_end=False and past_covariates starting before. "
                f"Expected {theorical_forecast_length}, got {len(forecasts[1])}",
            )

        # Past and future covariates
        for model_cls, kwargs, bounds, type in models_torch_cls_kwargs:
            if not type == "MixedCovariates":
                continue

            model = model_cls(
                random_state=0,
                **kwargs,
            )
            model.fit(
                self.ts_pass_train,
                past_covariates=self.ts_past_cov_train,
                future_covariates=self.ts_fut_cov_train,
            )

            forecasts = model.historical_forecasts(
                series=[self.ts_pass_val, self.ts_pass_val],
                past_covariates=[
                    self.ts_past_cov_valid_5_aft_start,
                    self.ts_past_cov_valid_same_start,
                ],
                future_covariates=[
                    self.ts_fut_cov_valid_7_aft_start,
                    self.ts_fut_cov_valid_16_bef_start,
                ],
                forecast_horizon=forecast_hrz,
                stride=1,
                retrain=True,
                overlap_end=False,
            )
            theorical_forecast_length = (
                self.ts_val_length
                - (bounds[0] + bounds[1])  # train sample length
                - (forecast_hrz - 1)  # if entire horizon is available, we can predict 1
                - 7  # future covs start 7 after target (more than past covs) -> shift
                - 2  # future covs in output chunk -> difference between horizon=10 and output_chunk_length=12
            )
            self.assertTrue(
                len(forecasts[0]) == theorical_forecast_length,
                f"Model {model_cls} does not return the right number of historical forecasts in case "
                f"of retrain=True and overlap_end=False and past_covariates and future_covariates with "
                f"different start. "
                f"Expected {theorical_forecast_length}, got {len(forecasts[0])}",
            )
            theorical_forecast_length = (
                self.ts_val_length
                - (bounds[0] + bounds[1])  # train sample length
                - (
                    forecast_hrz - 1
                )  # if entire horizon is available, we can predict 1,
                - 0  # all covs start at the same time as target -> no shift,
                - 2  # future covs in output chunk -> difference between horizon=10 and output_chunk_length=12
            )
            self.assertTrue(
                len(forecasts[1]) == theorical_forecast_length,
                f"Model {model_cls} does not return the right number of historical forecasts in case "
                f"of retrain=True and overlap_end=False and past_covariates with different start. "
                f"Expected {theorical_forecast_length}, got {len(forecasts[1])}",
            )

        # Future covariates only
        for model_cls, kwargs, bounds, type in models_torch_cls_kwargs:
            # todo case of DualCovariates (RNN)
            if type == "PastCovariates" or type == "DualCovariates":
                continue

            model = model_cls(
                random_state=0,
                **kwargs,
            )
            model.fit(self.ts_pass_train, future_covariates=self.ts_fut_cov_train)

            # Only fut covariate
            forecasts = model.historical_forecasts(
                series=[self.ts_pass_val, self.ts_pass_val],
                future_covariates=[
                    self.ts_fut_cov_valid_7_aft_start,
                    self.ts_fut_cov_valid_16_bef_start,
                ],
                forecast_horizon=forecast_hrz,
                stride=1,
                retrain=True,
                overlap_end=False,
            )

            self.assertTrue(
                len(forecasts) == 2,
                f"Model {model_cls} did not return a list of historical forecasts",
            )
            theorical_forecast_length = (
                self.ts_val_length
                - (bounds[0] + bounds[1])  # train sample length
                - (
                    forecast_hrz - 1
                )  # (horizon - 1): if entire horizon is available, we can predict 1,
                - 7  # future covs start 7 after target (more than past covs) -> shift
                - 2  # future covs in output chunk -> difference between horizon=10 and output_chunk_length=12
            )
            self.assertTrue(
                len(forecasts[0]) == theorical_forecast_length,
                f"Model {model_cls} does not return the right number of historical forecasts in case "
                f"of retrain=True and overlap_end=False and no past_covariates and future_covariates "
                f"with different start. "
                f"Expected {theorical_forecast_length}, got {len(forecasts[0])}",
            )
            theorical_forecast_length = (
                self.ts_val_length
                - (bounds[0] + bounds[1])  # train sample length
                - (forecast_hrz - 1)  # if entire horizon is available, we can predict 1
                - 0  # all covs start at the same time as target -> no shift
                - 2  # future covs in output chunk -> difference between horizon=10 and output_chunk_length=12
            )
            self.assertTrue(
                len(forecasts[1]) == theorical_forecast_length,
                f"Model {model_cls} does not return the right number of historical forecasts in case "
                f"of retrain=True and overlap_end=False and no past_covariates and future_covariates "
                f"with different start. "
                f"Expected {theorical_forecast_length}, got {len(forecasts[1])}",
            )

    def test_retrain(self):
        """test historical_forecasts for an untrained model with different retrain values."""

        def helper_hist_forecasts(retrain_val, start):
            model = LinearRegressionModel(lags=4, output_chunk_length=4)
            return model.historical_forecasts(
                self.ts_passengers, start=start, retrain=retrain_val, verbose=False
            )

        def retrain_f_invalid(
            counter, pred_time, train_series, past_covariates, future_covariates
        ):
            return False

        def retrain_f_missing_arg(
            counter, train_series, past_covariates, future_covariates
        ):
            if len(train_series) % 2 == 0:
                return True
            else:
                return False

        def retrain_f_invalid_ouput_int(
            counter, pred_time, train_series, past_covariates, future_covariates
        ):
            return 1

        def retrain_f_invalid_ouput_str(
            counter, pred_time, train_series, past_covariates, future_covariates
        ):
            return "True"

        def retrain_f_valid(
            counter, pred_time, train_series, past_covariates, future_covariates
        ):
            # only retrain once in first iteration
            if pred_time == pd.Timestamp("1959-09-01 00:00:00"):
                return True
            else:
                return False

        def retrain_f_delayed_true(
            counter, pred_time, train_series, past_covariates, future_covariates
        ):
            if counter > 1:
                return True
            else:
                return False

        # test callable
        helper_hist_forecasts(retrain_f_valid, 0.9)
        # missing the `pred_time` positional argument
        expected_msg = "the Callable `retrain` must have a signature/arguments matching the following positional"
        with pytest.raises(ValueError) as error_msg:
            helper_hist_forecasts(retrain_f_missing_arg, 0.9)
        self.assertTrue(str(error_msg.value).startswith(expected_msg))
        # returning a non-bool value (int)
        expected_msg = "Return value of `retrain` must be bool, received <class 'int'>"
        with pytest.raises(ValueError) as error_msg:
            helper_hist_forecasts(retrain_f_invalid_ouput_int, 0.9)
        self.assertTrue(str(error_msg.value).startswith(expected_msg))
        # returning a non-bool value (str)
        expected_msg = "Return value of `retrain` must be bool, received <class 'str'>"
        with pytest.raises(ValueError) as error_msg:
            helper_hist_forecasts(retrain_f_invalid_ouput_str, 0.9)
        self.assertTrue(str(error_msg.value).startswith(expected_msg))
        # predict fails but model could have been trained before the predict round
        expected_msg = "`retrain` is `False` in the first train iteration at prediction point (in time)"
        with pytest.raises(ValueError) as error_msg:
            helper_hist_forecasts(retrain_f_delayed_true, 0.9)
        self.assertTrue(str(error_msg.value).startswith(expected_msg))
        # always returns False, treated slightly different than `retrain=False` and `retrain=0`
        with pytest.raises(ValueError) as error_msg:
            helper_hist_forecasts(retrain_f_invalid, 0.9)
        self.assertTrue(str(error_msg.value).startswith(expected_msg))

        # test int
        helper_hist_forecasts(10, 0.9)
        expected_msg = "Model has not been fit before the first predict iteration at prediction point (in time)"
        # `retrain=0` with not-trained model, encountering directly a predictable time index
        with pytest.raises(ValueError) as error_msg:
            helper_hist_forecasts(0, 0.9)
        self.assertTrue(str(error_msg.value).startswith(expected_msg))

        # test bool
        helper_hist_forecasts(True, 0.9)
        # `retrain=False` with not-trained model, encountering directly a predictable time index
        expected_msg = "The model has not been fitted yet, and `retrain` is ``False``."
        with pytest.raises(ValueError) as error_msg:
            helper_hist_forecasts(False, 0.9)
        self.assertTrue(str(error_msg.value).startswith(expected_msg))

        expected_start = pd.Timestamp("1949-10-01 00:00:00")
        # start before first trainable time index should still work
        res = helper_hist_forecasts(True, pd.Timestamp("1949-09-01 00:00:00"))
        self.assertTrue(res.time_index[0] == expected_start)
        # start at first trainable time index should still work
        res = helper_hist_forecasts(True, expected_start)
        self.assertTrue(res.time_index[0] == expected_start)
        # start at last trainable time index should still work
        expected_end = pd.Timestamp("1960-12-01 00:00:00")
        res = helper_hist_forecasts(True, expected_end)
        self.assertTrue(res.time_index[0] == expected_end)
